/*
Copyright Rene Rivera 2008-2015
Copyright Franz Detro 2014
Distributed under the Boost Software License, Version 1.0.
(See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)
*/

#if !defined(BOOST_PREDEF_OS_H) || defined(BOOST_PREDEF_INTERNAL_GENERATE_TESTS)
#ifndef BOOST_PREDEF_OS_H
#define BOOST_PREDEF_OS_H
#endif

#include "os/aix.h"
#include "os/amigaos.h"
#include "os/android.h"
#include "os/beos.h"
#include "os/bsd.h"
#include "os/cygwin.h"
#include "os/haiku.h"
#include "os/hpux.h"
#include "os/irix.h"
#include "os/ios.h"
#include "os/linux.h"
#include "os/macos.h"
#include "os/os400.h"
#include "os/qnxnto.h"
#include "os/solaris.h"
#include "os/unix.h"
#include "os/vms.h"
#include "os/windows.h"

#endif
