#include <winuser.h>
#include <winver.h>

LANGUAGE 9, 1

1 ICON "@CMAKE_SOURCE_DIR@/win32/performous.ico"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION @VERSION_MAJOR@, @VERSION_MINOR@, @VERSION_PATCH@, @VERSION_TWEAK@
 PRODUCTVERSION @VERSION_MAJOR@, @VERSION_MINOR@, @VERSION_PATCH@, @VERSION_TWEAK@
 FILEFLAGSMASK VS_FFI_FILEFLAGSMASK
 FILEOS VOS_NT_WINDOWS32
 FILETYPE VFT_APP
 FILESUBTYPE VFT2_UNKNOWN
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904b0"
    BEGIN
      VALUE "CompanyName",  "@CMAKE_PROJECT_NAME@ Developers"
      VALUE "FileDescription",  "@CMAKE_PROJECT_NAME@ Game"
      VALUE "FileVersion", "@PROJECT_VERSION@"
      VALUE "InternalName", "Performous.exe"
      VALUE "LegalCopyright", "Copyright © 2009-2014 @CMAKE_PROJECT_NAME@ Team - GNU GPL v2 or later"
      VALUE "OriginalFilename", "Performous.exe"
      VALUE "ProductName", "@CMAKE_PROJECT_NAME@"
      VALUE "ProductVersion", "@PROJECT_VERSION@"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x409, 1200
  END
END
